"""Utility functions for python-bluezero."""
from enum import Enum
import logging
import inspect


def int_to_uint16(value_in):
    """
    Convert integer to Unsigned 16 bit little endian integer

    :param value_in: Integer < 65535 (0xFFFF)
    :return:
    """
    bin_string = '{:016b}'.format(value_in)
    big_byte = int(bin_string[0:8], 2)
    little_byte = int(bin_string[8:16], 2)
    return [little_byte, big_byte]


def sint16_to_int(byte_array):
    """
    Convert a signed 16-bit integer to integer

    :param byte_array:
    :return:
    """
    return int.from_bytes(byte_array, byteorder='little', signed=True)


def bytes_to_xyz(byte_array):
    """
    Split 6 byte long in integers representing x, y & z

    :param byte_array:
    :return:
    """
    x = sint16_to_int(byte_array[0:2]) / 1000  # pylint: disable=invalid-name
    y = sint16_to_int(byte_array[2:4]) / 1000  # pylint: disable=invalid-name
    z = sint16_to_int(byte_array[4:6]) / 1000  # pylint: disable=invalid-name

    return [x, y, z]


def int_to_uint32(value_in):
    """
    Convert integer to unsigned 32-bit (little endian)

    :param value_in:
    :return:
    """
    return list(value_in.to_bytes(4, byteorder='little', signed=False))


def bitwise_or_2lists(list1, list2):
    """
    Takes two bit patterns of equal length and performs the logical
    inclusive OR operation on each pair of corresponding bits
    """
    list_len = len(list2)
    return_list = [None] * list_len
    for i in range(list_len):
        return_list[i] = list1[i] | list2[i]
    return return_list


def bitwise_and_2lists(list1, list2):
    """
    Takes two bit patterns of equal length and performs the logical
    inclusive AND operation on each pair of corresponding bits
    """
    list_len = len(list2)
    return_list = [None] * list_len
    for i in range(list_len):
        return_list[i] = list1[i] & list2[i]
    return return_list


def bitwise_xor_2lists(list1, list2):
    """
    Takes two bit patterns of equal length and performs the logical
    inclusive XOR operation on each pair of corresponding bits
    """
    list_len = len(list1)
    return_list = [None] * list_len
    for i in range(list_len):
        return_list[i] = list1[i] ^ list2[i]
    return return_list


def url_to_advert(url, frame_type, tx_power):
    """
    Encode as specified
    https://github.com/google/eddystone/blob/master/eddystone-url/README.md

    :param url:
    :return:
    """
    prefix_sel = None
    prefix_start = None
    prefix_end = None
    suffix_sel = None
    suffix_start = None
    suffix_end = None

    prefix = ('http://www.', 'https://www.', 'http://', 'https://')

    suffix = ('.com/', '.org/', '.edu/', '.net/', '.info/', '.biz/', '.gov/',
              '.com', '.org', '.edu', '.net', '.info', '.biz', '.gov'
              )
    encode_search = True

    for domain in prefix:
        if domain in url and encode_search is True:
            # print('match prefix ' + url)
            prefix_sel = prefix.index(domain)
            prefix_start = url.index(prefix[prefix_sel])
            prefix_end = len(prefix[prefix_sel]) + prefix_start
            encode_search = False

    encode_search = True
    for tld in suffix:
        if tld in url and encode_search is True:
            # print('match suffix ' + y)
            suffix_sel = suffix.index(tld)
            suffix_start = url.index(suffix[suffix_sel])
            suffix_end = len(suffix[suffix_sel]) + suffix_start
            encode_search = False

    service_data = [frame_type]
    service_data.extend([tx_power])
    if suffix_start is None:
        suffix_start = len(url)
        service_data.extend([prefix_sel])
        for domain in range(prefix_end, suffix_start):
            service_data.extend([ord(url[domain])])
    elif suffix_end == len(url):
        service_data.extend([prefix_sel])
        for domain in range(prefix_end, suffix_start):
            service_data.extend([ord(url[domain])])
        service_data.extend([suffix_sel])
    else:
        service_data.extend([prefix_sel])
        for domain in range(prefix_end, suffix_start):
            service_data.extend([ord(url[domain])])
        service_data.extend([suffix_sel])
        for domain in range(suffix_end, len(url)):
            service_data.extend([ord(url[domain])])

    return service_data


def get_fn_parameters(fn):
    """ return the number of input parameters of the fn , None on error"""
    param_len = len(inspect.getfullargspec(fn).args)
    if inspect.ismethod(fn):
        param_len -= 1
    return param_len


def create_module_logger(module_name):
    """helper function to create logger in Bluezero modules"""
    logger = logging.getLogger(module_name)
    strm_hndlr = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    strm_hndlr.setFormatter(formatter)
    logger.addHandler(strm_hndlr)
    return logger

# Improve the above logger helper so that report level can be easily changed
# by users. This is not ready for inclusion but leaving it here in case anyone
# has some ideas
#
# class DebugLevels(Enum):
#     CRITICAL = 50
#     ERROR = 40
#     WARNING = 30
#     INFO = 20
#     DEBUG = 10
#     NOTSET = 0
#
#
# class BluezeroLogger:
#     def __init__(self, module_name):
#         self.logger = logging.getLogger(module_name)
#         ch = logging.StreamHandler()
#         formatter = logging.Formatter(
#             '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
#         ch.setFormatter(formatter)
#         self.logger.addHandler(ch)
#         return self.logger
#
#     @property
#     def level(self):
#         return DebugLevels(self.logger.getEffectiveLevel()).name
#
#     @level.setter
#     def level(self, logger_level):
#         self.logger.setLevel()



"""Access BlueZ Media Player functionality"""
import dbus

# python-bluezero imports
from bluezero import constants
from bluezero import dbus_tools
from bluezero import tools

logger = tools.create_module_logger(__name__)


class MediaPlayerError(Exception):
    """Custom exception"""
    pass


def _find_player_path(device_address):
    """ finds the player_path corresponding to the device addr"""
    player_path_list = []
    mngd_objs = dbus_tools.get_managed_objects()
    for path in dbus_tools.get_managed_objects():
        if mngd_objs[path].get(constants.MEDIA_PLAYER_IFACE):
            player_path_list.append(path)

    for path in player_path_list:
        found_address = dbus_tools.get_device_address_from_dbus_path(path)
        if device_address == found_address:
            return path
    raise MediaPlayerError("No player found for the device")


class MediaPlayer:
    """Bluetooth MediaPlayer Class.
    This class instantiates an object that is able to interact with
    the player of a Bluetooth device and get audio from its source.
    """

    def __init__(self, device_addr):
        """Default initialiser.

        Creates the interface to the remote Bluetooth device.

        :param device_addr: Address of Bluetooth device player to use.
        """
        self.player_path = _find_player_path(device_addr)
        self.player_object = dbus_tools.get_dbus_obj(self.player_path)
        self.player_methods = dbus_tools.get_dbus_iface(
            constants.MEDIA_PLAYER_IFACE, self.player_object)
        self.player_props = dbus_tools.get_dbus_iface(
            dbus.PROPERTIES_IFACE, self.player_object)

    @property
    def browsable(self):
        """If present indicates the player can be browsed using MediaFolder
        interface."""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Browsable')

    @property
    def searchable(self):
        """If present indicates the player can be searched using
        MediaFolder interface."""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Searchable')

    @property
    def track(self):
        """Return a dict of the track metadata."""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Track')

    @property
    def device(self):
        """Return Device object path"""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Device')

    @property
    def playlist(self):
        """Return the Playlist object path."""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Playlist')

    @property
    def equalizer(self):
        """Return the equalizer value."""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Equalizer')

    @equalizer.setter
    def equalizer(self, value):
        """Possible values: "off" or "on"."""
        self.player_props.Set(
            constants.MEDIA_PLAYER_IFACE, 'Equalizer', value)

    @property
    def name(self):
        """Return the player name"""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Name')

    @property
    def repeat(self):
        """Return the repeat value"""
        return self.player_props.Get(
            constants.MEDIA_PLAYER_IFACE, 'Repeat')

    @repeat.setter
    def repeat(self, value):
        """Possible values: "off", "singletrack", "alltracks" or "group"""
        self.player_props.Set(
            constants.MEDIA_PLAYER_IFACE, 'Repeat', value)

    @property
    def shuffle(self):
        """Return the shuffle value"""
        return self.player_props.Get(constants.MEDIA_PLAYER_IFACE, 'Shuffle')

    @shuffle.setter
    def shuffle(self, value):
        """"Possible values: "off", "alltracks" or "group" """
        self.player_props.Set(constants.MEDIA_PLAYER_IFACE, 'Shuffle', value)

    @property
    def status(self):
        """Return the status of the player
        Possible status: "playing", "stopped", "paused",
        "forward-seek", "reverse-seek" or "error" """
        return self.player_props.Get(constants.MEDIA_PLAYER_IFACE, 'Status')

    @property
    def subtype(self):
        """Return the player subtype"""
        return self.player_props.Get(constants.MEDIA_PLAYER_IFACE, 'Subtype')

    def type(self, player_type):
        """Player type. Possible values are:

                * "Audio"
                * "Video"
                * "Audio Broadcasting"
                * "Video Broadcasting"
        """
        self.player_props.Set(
            constants.MEDIA_PLAYER_IFACE, 'Type', player_type)

    @property
    def position(self):
        """Return the playback position in milliseconds."""
        return self.player_props.Get(constants.MEDIA_PLAYER_IFACE, 'Position')

    def next(self):
        """Goes the next track and play it."""
        self.player_methods.Next()

    def play(self):
        """Resume the playback."""
        self.player_methods.Play()

    def pause(self):
        """Pause the track."""
        self.player_methods.Pause()

    def stop(self):
        """Stops the playback."""
        self.player_methods.Stop()

    def previous(self):
        """Goes the previous track and play it"""
        self.player_methods.Previous()

    def fast_forward(self):
        """Fast forward playback, this action is only stopped
        when another method in this interface is called.
        """
        self.player_methods.FastForward()

    def rewind(self):
        """Rewind playback, this action is only stopped
        when another method in this interface is called.
        """
        self.player_methods.Rewind()



"""
Tests for the Adapter.

This test class makes use of sample DBus data captured from a live device,
imported as ``tests.obj_data``.
"""
import sys
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants


def mock_get_all(iface):
    """
    Mock the DBus ``GetAll()`` operation for an interface.

    :param string iface: DBus interface to ``GetAll()`` on
                         (e.g. ``'org.bluez.Adapter1'``)
    :return: DBus Interface Properties
    :rtype: Dictionary
    """
    return tests.obj_data.full_ubits['/org/bluez/hci0'][iface]


def mock_get(iface, prop):
    """
    Mock the DBus ``Get()`` operation for a property on an interface.

    :param string iface: DBus interface, e.g. ``'org.bluez.Adapter1'``
    :param string prop: DBus property name, e.g. ``'Name'``
    :return: DBus property
    :rtype: *property dependent*
    """
    return mock_get_all(iface)[prop]


def mock_set(iface, prop, value):
    """
    Mock the DBus ``Set()`` operation for a property on an interface.

    :param string iface: DBus interface, e.g. ``'org.bluez.Adapter1'``
    :param string prop: DBus property name, e.g. ``'Name'``
    :param value: Value to set the DBus property to
    :return: None
    """
    tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop] = value


class TestBluezeroAdapter(unittest.TestCase):
    """
    Mock a BLE Adapter.
    """

    def setUp(self):
        """
        Patch the DBus module using ``MagicMock`` from ``unittest.mock``

        :return: None

        Three modules are mocked and patched into the system:

        #. DBus

           * Calls to ``GetManagedObjects()``, ``Get()``, ``Set()``, and
             ``GetAll()`` are mocked using a captured DBus dictionary.
        #. DBus Mainloop
        #. GObject
        """
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }

        dbus_mock_iface = self.dbus_mock.Interface.return_value
        dbus_mock_iface.GetManagedObjects.return_value = \
            tests.obj_data.full_ubits
        dbus_mock_iface.Get = mock_get
        dbus_mock_iface.Set = mock_set
        dbus_mock_iface.GetAll = mock_get_all

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()

        from bluezero import adapter
        self.module_under_test = adapter
        # self.adapter_device = 'hci0'  # Currently unused
        self.adapter_name = 'linaro-alip'
        self.path = '/org/bluez/hci0'

    def tearDown(self):
        """
        Stop the module patching.
        """
        self.module_patcher.stop()

    def test_list_adapters(self):
        """
        Test ``Adapter.list_adapters()``
        """
        adapters = self.module_under_test.list_adapters()
        self.assertListEqual(['00:00:00:00:5A:AD'], adapters)

    def test_adapter_address(self):
        """
        Test the adapter ``address`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        self.assertEqual(dongle.address, '00:00:00:00:5A:AD')

    def test_adapter_default(self):
        """
        Test the default ``Adapter()`` instantation gets the correct address.
        """
        dongle = self.module_under_test.Adapter()
        self.assertEqual(dongle.address, '00:00:00:00:5A:AD')

    def test_get_all(self):
        """
        Test the ``get_all()`` method for retrieving all the DBus properties.
        """
        dongle = self.module_under_test.Adapter(self.path)
        self.assertEqual(dongle.get_all(),
                         mock_get_all(constants.ADAPTER_INTERFACE))

    def test_adapter_name(self):
        """
        Test the adapter ``name`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.name, self.adapter_name)

    def test_adapter_alias(self):
        """
        Test the adapter ``alias`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.alias, self.adapter_name)

    def test_adapter_alias_write(self):
        """
        Test that an adapter ``alias`` can be set.
        """
        dev_name = 'my-test-dev'
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.alias = dev_name
        self.assertEqual(dongle.alias, dev_name)

    def test_class(self):
        """
        Test the adapter ``bt_class`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.bt_class, 4980736)

    def test_adapter_power(self):
        """
        Test the adapter ``powered`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.powered, 1)

    def test_adapter_power_write(self):
        """
        Test that the adapter ``powered`` property can be set.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.powered = 0
        self.assertEqual(dongle.powered, False)

    def test_adapter_discoverable(self):
        """
        Test the adapter ``discoverable`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.discoverable, False)

    def test_adapter_discoverable_write(self):
        """
        Test that the adapter ``discoverable`` property can be set.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.discoverable = 1
        self.assertEqual(dongle.discoverable, True)

    def test_adapter_pairable(self):
        """
        Test the adapter ``pairable`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.pairable, 1)

    def test_adapter_pairable_write(self):
        """
        Test that the adapter ``pairable`` property can be set.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.pairable = 0
        self.assertEqual(dongle.pairable, 0)

    def test_adapter_pairabletimeout(self):
        """
        Test the adapter ``pairabletimeout`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.pairabletimeout, 0)

    def test_adapter_pairabletimeout_write(self):
        """
        Test that the adapter ``pairabletimeout`` property can be set.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.pairabletimeout = 220
        self.assertEqual(dongle.pairabletimeout, 220)

    def test_adapter_discoverable_timeout(self):
        """
        Test the adapter ``discoverabletimeout`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.discoverabletimeout, 180)

    def test_adapter_discoverabletimeout_write(self):
        """
        Test that the adapter ``discoverabletimeout`` property can be set.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.discoverabletimeout = 220
        self.assertEqual(dongle.discoverabletimeout, 220)

    def test_adapter_discovering(self):
        """
        Test the adapter ``discovering`` property.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        self.assertEqual(dongle.discovering, False)

    @unittest.skip('mock of discovery not implemented')
    def test_start_discovery(self):
        """
        Test the adapter ``nearby_discovering()`` method.
        """
        dongle = self.module_under_test.Adapter(self.path)
        # test
        dongle.nearby_discovery()
        self.assertEqual(dongle.discovering, 1)


if __name__ == '__main__':
    # avoid writing to stderr
    unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout,
                                                     verbosity=2))



import dbus
import dbusmock
from pathlib import Path
import subprocess
from unittest import mock
from tests.mock_eventloop import MockAsync, run_pending_events

# Module under test
from examples import cpu_temperature


class TestAdapterExample(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')

    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_create_peripheral(self):
        # Test currently has limited value because the RegisterApplication
        # does not publish the peripheral with dbusmock
        # (dbusmock_templates/bluez_scan.py) so further testing cannot happen.

        # Add adapter
        path0 = self.dbusmock_bluez.AddAdapter('hci0', 'My-Peripheral')
        path1 = self.dbusmock_bluez.AddAdapter('hci1', 'My-Adapter')

        # Setup to monitor calls
        bus = self.get_dbus(True)
        peri_obj = bus.get_object('org.bluez', path1)
        peri_mock = dbus.Interface(peri_obj, dbusmock.MOCK_IFACE)
        adptr_obj = bus.get_object('org.bluez', path0)
        adptr_mock = dbus.Interface(adptr_obj, dbusmock.MOCK_IFACE)
        # Create
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            cpu_temperature.main(adapter_address='00:01:02:03:04:05')
        device_calls = list(str(call[1]) for call in peri_mock.GetCalls())
        adptr_calls = list(str(call[1]) for call in adptr_mock.GetCalls())
        self.assertEqual(['RegisterApplication'], adptr_calls)
        self.assertEqual([], device_calls)




import dbus
import dbusmock
import io
import logging
from pathlib import Path
import subprocess
from unittest import mock
from tests.mock_eventloop import MockAsync

# Module under test
from examples import microbit_uart


class TestAdapterExample(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')

    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_run_uart_example(self):
        # Horrible hardcoded path that refers to characteristic defined in
        # microbit_data. Needs refactoring but this was quick to prove it
        # worked as a test methodology
        tx_path = '/org/bluez/hci0/dev_DD_02_02_02_02_02/service0028/char002b'
        rx_path = '/org/bluez/hci0/dev_DD_02_02_02_02_02/service0028/char0029'
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        path = self.dbusmock_bluez.AddDevice('hci0',
                                             'DD:02:02:02:02:02',
                                             'micro:bit[uart]')
        bus = self.get_dbus(True)
        dev_obj = bus.get_object('org.bluez', path)
        dev_mock = dbus.Interface(dev_obj, dbusmock.MOCK_IFACE)
        rx_obj = bus.get_object('org.bluez', rx_path)
        rx_mock = dbus.Interface(rx_obj, dbusmock.MOCK_IFACE)
        tx_obj = bus.get_object('org.bluez', tx_path)
        tx_mock = dbus.Interface(tx_obj, dbusmock.MOCK_IFACE)
        with mock.patch('bluezero.async_tools.EventLoop', MockAsync):
            with mock.patch('sys.stdout', new=io.StringIO()) as fake_out:
                microbit_uart.main('00:01:02:03:04:05', 'DD:02:02:02:02:02')

        # Check device calls
        device_calls = list(str(call[1]) for call in dev_mock.GetCalls())
        self.assertEqual(['Connect', 'Disconnect'], device_calls)
        # Check UART RX calls
        rx_calls = list(str(call[1]) for call in rx_mock.GetCalls())
        self.assertEqual(['WriteValue', 'WriteValue'], rx_calls)
        # Check UART TX calls
        tx_calls = list(str(call[1]) for call in tx_mock.GetCalls())
        self.assertEqual(['StartNotify'], tx_calls)
        # Check notification value is printed to screen
        self.assertEqual('Ping#\n', fake_out.getvalue())





"""
Code for create a Bluetooth Low Energy Observer (beacon scanner) application
"""
from collections import namedtuple
import uuid
import dbus
from bluezero import adapter
from bluezero import async_tools
from bluezero import tools

logger = tools.create_module_logger(__name__)
packet_callback = None

DEVICE_INTERFACE = 'org.bluez.Device1'
EDDYSTONE_SRV_UUID = '0000feaa-0000-1000-8000-00805f9b34fb'

EddyURL = namedtuple('EddyURL', ['url', 'tx_pwr', 'rssi'])
EddyUID = namedtuple('EddyUID', ['namespace', 'instance', 'tx_pwr', 'rssi'])
EddyTLM = namedtuple('EddyTLM',
                     ['version', 'battery', 'temperature', 'count', 'uptime',
                      'tx_pwr', 'rssi'])
iBeacon = namedtuple('iBeacon',  # pylint: disable=invalid-name
                     ['UUID', 'major', 'minor', 'tx_pwr', 'rssi'])
AltBeacon = namedtuple('AltBeacon',
                       ['UUID', 'major', 'minor', 'tx_pwr', 'rssi'])


class Scanner:
    """
    For scanning of Bluetooth Low Energy (BLE) beacons
    """
    remove_list = set()
    mainloop = async_tools.EventLoop()
    on_eddystone_url = None
    on_eddystone_uid = None
    on_eddystone_tlm = None
    on_ibeacon = None
    on_altbeacon = None

    @classmethod
    def start_event_loop(cls):
        """
        Start the event loop
        """
        cls.mainloop.run()

    @classmethod
    def stop_scan(cls):
        """
        Stop scanning for beacons
        """
        cls.mainloop.quit()
        cls.dongle.stop_discovery()

    @classmethod
    def clean_beacons(cls):
        """
        Remove beacons from BlueZ's `devices` list so every advert from a
        beacon is reported
        """
        not_found = set()
        for rm_dev in cls.remove_list:
            logger.debug('Remove %s', rm_dev)
            try:
                cls.dongle.remove_device(rm_dev)
            except dbus.exceptions.DBusException as dbus_err:
                if dbus_err.get_dbus_name() == 'org.bluez.Error.DoesNotExist':
                    not_found.add(rm_dev)
                else:
                    raise dbus_err
        for lost in not_found:
            cls.remove_list.remove(lost)

    @classmethod
    def process_eddystone(cls, data, rssi):
        """
        Extract Eddystone data from advertisement service data

        :param data: Bytes from Service Data in advertisement
        :param rssi: Received Signal Strength value
        """
        _url_prefix_scheme = ['http://www.', 'https://www.',
                              'http://', 'https://', ]
        _url_encoding = ['.com/', '.org/', '.edu/', '.net/', '.info/',
                         '.biz/', '.gov/', '.com', '.org', '.edu',
                         '.net', '.info', '.biz', '.gov']
        tx_pwr = int.from_bytes([data[1]], 'big', signed=True)
        if data[0] == 0x00:
            namespace_id = int.from_bytes(data[2:12], 'big')
            instance_id = int.from_bytes(data[12:18], 'big')
            logger.info('\t\tEddystone UID: %s - %s \u2197 %s', namespace_id,
                        instance_id, tx_pwr)
            data = EddyUID(namespace_id, instance_id, tx_pwr, rssi)
            if cls.on_eddystone_uid:
                cls.on_eddystone_uid(data)
        elif data[0] == 0x10:
            prefix = data[2]
            encoded_url = data[3:]
            full_url = _url_prefix_scheme[prefix]
            for letter in encoded_url:
                if letter < len(_url_encoding):
                    full_url += _url_encoding[letter]
                else:
                    full_url += chr(letter)
            logger.info('\t\tEddystone URL: %s  \u2197 %s  \u2198 %s',
                        full_url, tx_pwr, rssi)
            data = EddyURL(url=full_url, tx_pwr=tx_pwr, rssi=int(rssi))
            if cls.on_eddystone_url:
                cls.on_eddystone_url(data)
        elif data[0] == 0x20:
            version = int(data[1])
            # Only support plain beacons ATM
            if version == 0:
                voltage = int.from_bytes(
                    data[2:4], byteorder='big', signed=False)
                temperature = int.from_bytes(
                    data[4:6], byteorder='big', signed=True)/256.0
                count = int.from_bytes(
                    data[6:10], byteorder='big', signed=False)
                time = int.from_bytes(
                    data[10:], byteorder='big', signed=False) / 10
                logger.info('\t\tEddystone TLM: %s, %s mV, %s C, %s, %s s '
                            '\u2197 %s  \u2198 %s',
                            version, voltage, temperature, count, time,
                            tx_pwr, rssi)
                data = EddyTLM(version, voltage, temperature, count, time,
                               tx_pwr, rssi)
                if cls.on_eddystone_tlm:
                    cls.on_eddystone_tlm(data)

    @classmethod
    def process_ibeacon(cls, data, rssi, beacon_type='iBeacon'):
        """
        Extract iBeacon or AltBeacon data from Manufacturer Data in an
        advertisement

        :param data: Bytes from manufacturer data
        :param rssi: Received Signal Strength value
        :param beacon_type: iBeacon or AltBeacon
        """
        beacon_uuid = uuid.UUID(bytes=bytes(data[2:18]))
        major = int.from_bytes(bytearray(data[18:20]), 'big', signed=False)
        minor = int.from_bytes(bytearray(data[20:22]), 'big', signed=False)
        tx_pwr = int.from_bytes([data[22]], 'big', signed=True)
        logger.info('%s: %s - %s - %s  \u2197 %s \u2198 %s', beacon_type,
                    beacon_uuid, major, minor, tx_pwr, rssi)
        if beacon_type == 'iBeacon' and cls.on_ibeacon:
            data = iBeacon(beacon_uuid, major, minor, tx_pwr, rssi)
            cls.on_ibeacon(data)
        elif beacon_type == 'AltBeacon' and cls.on_altbeacon:
            data = AltBeacon(beacon_uuid, major, minor, tx_pwr, rssi)
            cls.on_altbeacon(data)

    @staticmethod
    def ble_16bit_match(uuid_16, srv_data):
        """
        Utility method to test 16 bit UUID against Bluetooth SIG 128 bit UUID
        used in service data

        :param uuid_16: 16 Bit UUID value
        :param srv_data:
        :return:
        """
        uuid_128 = f'0000{uuid_16}-0000-1000-8000-00805f9b34fb'
        return uuid_128 == list(srv_data.keys())[0]

    @classmethod
    def on_device_found(cls, bz_device_obj):
        """
        Callback to look at BLE advertisement to see if it is a recognised
        beacon and if it is, then call the relevant processing function

        :param bz_device_obj: Bluezero device object of discovered device
        """
        rssi = bz_device_obj.RSSI
        service_data = bz_device_obj.service_data
        manufacturer_data = bz_device_obj.manufacturer_data
        if service_data and cls.ble_16bit_match('feaa', service_data):
            cls.process_eddystone(service_data[EDDYSTONE_SRV_UUID],
                                  rssi)
            cls.remove_list.add(str(bz_device_obj.remote_device_path))
        elif manufacturer_data:
            for mfg_id in manufacturer_data:
                # iBeacon 0x004c
                if mfg_id == 0x004c and manufacturer_data[mfg_id][0] == 0x02:
                    cls.process_ibeacon(manufacturer_data[mfg_id], rssi)
                    cls.remove_list.add(str(bz_device_obj.remote_device_path))
                # AltBeacon 0xacbe
                elif all((mfg_id == 0xffff,
                          manufacturer_data[mfg_id][0:2] == [0xbe, 0xac])):
                    cls.process_ibeacon(manufacturer_data[mfg_id], rssi,
                                        beacon_type='AltBeacon')
                    cls.remove_list.add(str(bz_device_obj.remote_device_path))
                # elif mfg_id == 0x0310:
                #     print(f'\t\tBlue Maestro {manufacturer_data[mfg_id]}')
                #     remove_list.add(device_path)
        cls.clean_beacons()

    @classmethod
    def start_beacon_scan(cls,
                          on_eddystone_url=None,
                          on_eddystone_uid=None,
                          on_eddystone_tlm=None,
                          on_ibeacon=None,
                          on_altbeacon=None):
        """
        Start scan for beacons. Provided callback will be called if matching
        beacon type is found.
        All callbacks take one argument which is a named tuple with the fields
        relevant for that format.

        - Eddystone URL = ['url', 'tx_pwr', 'rssi']
        - Eddystone UID = ['namespace', 'instance', 'tx_pwr', 'rssi']
        - Eddystone TLM = ['version', 'voltage', 'temperature', 'count',
                           'time', 'tx_pwr', 'rssi']
        - iBeacon = ['UUID', 'major', 'minor', 'tx_pwr', 'rssi']
        - AltBeacon = ['UUID', 'major', 'minor', 'tx_pwr', 'rssi']

        :param on_eddystone_url: Callback for Eddystone URL format
        :param on_eddystone_uid: Callback for Eddystone UID format
        :param on_eddystone_tlm: Callback for Eddystone TLM format
        :param on_ibeacon: Callback for iBeacon format
        :param on_altbeacon: Callback for AltBeacon format
        """
        cls.dongle = adapter.Adapter()
        cls.on_eddystone_url = on_eddystone_url
        cls.on_eddystone_uid = on_eddystone_uid
        cls.on_eddystone_tlm = on_eddystone_tlm
        cls.on_ibeacon = on_ibeacon
        cls.on_altbeacon = on_altbeacon

        cls.dongle.on_device_found = cls.on_device_found

        cls.dongle.show_duplicates()
        cls.dongle.start_discovery()
        try:
            cls.start_event_loop()
        except KeyboardInterrupt:
            cls.stop_scan()


def scan_eddystone(on_data=None):
    """
    Provide a callback for 'on_data'. The callback will be run whenever
    an Eddystone URL packet is detected.

    :param on_data: A function to be called on Eddystone packet
    :return: None
    """
    try:
        Scanner.start_beacon_scan(on_eddystone_url=on_data)
    except KeyboardInterrupt:
        Scanner.stop_scan()



"""Global constants file for the python-bluezero project.

This file is a single location for the different object paths that are used as
constants around the python-bluezero library.

:Example:

.. code-block:: python

   from bluezero import constants
   my_function(constants.CONST_1, constants.CONST_2)

"""

# General D-Bus Object Paths
#: The DBus Object Manager interface
DBUS_OM_IFACE = 'org.freedesktop.DBus.ObjectManager'
#: DBus Properties interface
DBUS_PROP_IFACE = 'org.freedesktop.DBus.Properties'

# General Bluez D-Bus Object Paths
#: BlueZ DBus Service Name
BLUEZ_SERVICE_NAME = 'org.bluez'
#: BlueZ DBus adapter interface
ADAPTER_INTERFACE = 'org.bluez.Adapter1'
#: BlueZ DBus device Interface
DEVICE_INTERFACE = 'org.bluez.Device1'

# Bluez GATT D-Bus Object Paths
#: BlueZ DBus GATT manager Interface
GATT_MANAGER_IFACE = 'org.bluez.GattManager1'
#: BlueZ DBus GATT Profile Interface
GATT_PROFILE_IFACE = 'org.bluez.GattProfile1'
#: BlueZ DBus GATT Service Interface
GATT_SERVICE_IFACE = 'org.bluez.GattService1'
#: BlueZ DBus GATT Characteristic Interface
GATT_CHRC_IFACE = 'org.bluez.GattCharacteristic1'
#: BlueZ DBus GATT Descriptor Interface
GATT_DESC_IFACE = 'org.bluez.GattDescriptor1'

# Bluez Advertisment D-Bus object paths
#: BlueZ DBus Advertising Manager Interface
LE_ADVERTISING_MANAGER_IFACE = 'org.bluez.LEAdvertisingManager1'
#: BlueZ DBus Advertisement Interface
LE_ADVERTISEMENT_IFACE = 'org.bluez.LEAdvertisement1'

# Bluez Media D-Bus object paths
#: BlueZ DBus Media player Interface
MEDIA_PLAYER_IFACE = 'org.bluez.MediaPlayer1'

# Bluezero local D-Bus publish location
#: Bluezero D-Bus Name
BLUEZERO_DBUS_NAME = 'ukBaz.bluezero'
#: Bluezero D-Bus object root
BLUEZERO_DBUS_OBJECT = '/ukBaz/bluezero'



import sys
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants


def mock_get(iface, prop):
    return tests.obj_data.full_ubits['/org/bluez/hci0/dev_D4_AE_95_4C_3E_A4'][iface][prop]


def mock_set(iface, prop, value):
    tests.obj_data.full_ubits['/org/bluez/hci0/dev_D4_AE_95_4C_3E_A4'][iface][prop] = value


class TestBluezeroDevice(unittest.TestCase):

    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.dbus_mock.Interface.return_value.Get = mock_get
        self.dbus_mock.Interface.return_value.Set = mock_set
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import device
        self.module_under_test = device
        self.path = '/org/bluez/hci0/dev_D4_AE_95_4C_3E_A4'
        self.adapter_path = '/org/bluez/hci0'
        self.dev_name = 'BBC micro:bit [zezet]'
        self.adapter_addr = '00:00:00:00:5A:AD'
        self.device_addr = 'D4:AE:95:4C:3E:A4'

    def tearDown(self):
        self.module_patcher.stop()

    def test_bad_device(self):
        self.assertRaises(ValueError,
                          self.module_under_test.Device,
                          self.adapter_addr,
                          "bad_device_address")

    def test_device_name(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.name, self.dev_name)

    def test_connected(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        conn_state = ble_dev.connected
        self.assertEqual(conn_state, False)
        """
        self.dbusmock_bluez.ConnectDevice(self.adapter_name,
                                          self.address)
        conn_state = self.ble_dev.connected
        self.assertEqual(conn_state, True)
    """
    def test_address(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.address, self.device_addr)

    def test_name(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.name, self.dev_name)

    @unittest.skip('Do not know value to use for icon')
    def test_icon(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.icon, 'dave')

    @unittest.skip('Do not know value to use for class')
    def test_class(self):
        pass

    def test_appearance(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.appearance, 0x0200)

    def test_uuids(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.uuids,
                         ['00001800-0000-1000-8000-00805f9b34fb',
                          '00001801-0000-1000-8000-00805f9b34fb',
                          '0000180a-0000-1000-8000-00805f9b34fb',
                          'e95d0753-251d-470a-a062-fa1922dfa9a8',
                          'e95d127b-251d-470a-a062-fa1922dfa9a8',
                          'e95d6100-251d-470a-a062-fa1922dfa9a8',
                          'e95d9882-251d-470a-a062-fa1922dfa9a8',
                          'e95dd91d-251d-470a-a062-fa1922dfa9a8',
                          'e95df2d8-251d-470a-a062-fa1922dfa9a8'])

    def test_paired(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.paired, False)

    def test_trusted(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.trusted, False)

    def test_blocked(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.blocked, False)

    def test_alias(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.alias, self.dev_name)

    def test_adapter(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev._adapter, self.adapter_path)

    def test_legacy_pairing(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.legacy_pairing, False)

    @unittest.skip('Do not know value to use for modalias')
    def test_modalias(self):
        pass

    @unittest.skip('No value from BlueZ for RSSI')
    def test_rssi(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.RSSI, -79)

    @unittest.skip('No value from BlueZ for Tx Power')
    def test_txpower(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.tx_power, 0)

    @unittest.skip('No test data for Manufacturer Data')
    def test_manufacturer_data(self):
        pass

    @unittest.skip('No test data for Service Data')
    def test_service_data(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)

    def test_services_resolved(self):
        ble_dev = self.module_under_test.Device(self.adapter_addr, self.device_addr)
        self.assertEqual(ble_dev.services_resolved, False)

    @unittest.skip('Not in BlueZ 5.42')
    def test_adverting_flags(self):
        pass

if __name__ == '__main__':
    # avoid writing to stderr
    unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout,
                                                     verbosity=2))



"""
Level 1 file for creating Eddystone beacons

Eddystone is a Bluetooth Low Energy beacon profile released by
Google in July 2015
https://github.com/google/eddystone

This is the broadcaster role which currently requires BlueZ to
have the experimental flag enabled
"""
from bluezero import tools
from bluezero import broadcaster


class EddystoneURL:
    """
    Create and start broadcasting a Eddystone URL beacon

    The Eddystone-URL frame broadcasts a URL using a compressed encoding
    format in order to fit more within the limited advertisement packet.
    :Example:

    >>> from bluezero import eddystone_beacon
    >>> eddystone_beacon.EddystoneURL('https://github.com/ukBaz')


    :param url: String containing URL e.g. ('http://camjam.me')
    :param tx_power: Value of Tx Power of advertisement (Not implemented)


    """
    def __init__(self, url, tx_power=0x08):
        """

        :param url: String containing URL e.g. ('http://camjam.me')
        :param tx_power: Value of Tx Power of advertisement (Not implemented)

        """
        service_data = tools.url_to_advert(url, 0x10, tx_power)
        if len(service_data) > 20:
            raise Exception('URL too long')
        url_beacon = broadcaster.Beacon()
        url_beacon.add_service_data('FEAA', service_data)
        url_beacon.start_beacon()



"""Class and methods that represent a remote Bluetooth Device.

Classes:

- Device -- Remote Bluetooth Device Class
"""
import dbus
import dbus.exceptions

from bluezero import constants
from bluezero import dbus_tools
import bluezero.adapter
from bluezero import tools


logger = tools.create_module_logger(__name__)


class Device:
    """Remote Bluetooth Device Class.

    This class instantiates an object that interacts with a remote
    Bluetooth device.
    """

    @staticmethod
    def available(adapter_address=None):
        """A generator yielding a Device object for every discovered device."""
        mng_objs = dbus_tools.get_managed_objects()
        adapters = {
            adapter.path: adapter.address
            for adapter in bluezero.adapter.Adapter.available()
        }
        for obj in mng_objs.values():
            device = obj.get(constants.DEVICE_INTERFACE, None)
            if device:
                adapter = adapters[device['Adapter']]
                if adapter_address is None or adapter == adapter_address:
                    yield Device(adapter, device['Address'])

    def __init__(self, adapter_addr, device_addr):
        """Default initialiser.

        Creates object for the specified remote Bluetooth device.
        This is on the specified adapter specified.

        :param adapter_addr: Address of the local Bluetooth adapter.
        :param device_addr: Address of the remote Bluetooth device.
        """
        self.bus = dbus.SystemBus()
        device_path = dbus_tools.get_dbus_path(adapter_addr, device_addr)
        if not device_path:
            raise ValueError("Cannot find a device: " + device_addr +
                             " using adapter: " + adapter_addr)

        self.remote_device_path = device_path
        self.remote_device_obj = self.bus.get_object(
            constants.BLUEZ_SERVICE_NAME,
            self.remote_device_path)
        self.remote_device_methods = dbus.Interface(
            self.remote_device_obj,
            constants.DEVICE_INTERFACE)
        self.remote_device_props = dbus.Interface(self.remote_device_obj,
                                                  dbus.PROPERTIES_IFACE)

    @property
    def address(self):
        """Return the remote device address."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Address')

    @property
    def name(self):
        """Return the remote device name."""
        return dbus_tools.get(
            self.remote_device_props, constants.DEVICE_INTERFACE,
            'Name', None)

    @property
    def icon(self):
        """
        Proposed icon name.

        This is set according to the freedesktop.org icon naming specification.
        """
        return dbus_tools.get(
            self.remote_device_props, constants.DEVICE_INTERFACE, 'Icon', None)

    @property
    def bt_class(self):
        """The Bluetooth class of device of the remote device."""
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'Class', None)

    @property
    def appearance(self):
        """External appearance of device, as found on GAP service."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Appearance')

    @property
    def uuids(self):
        """List of 128-bit UUIDs that represent available remote services."""
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'UUIDs', None)

    @property
    def paired(self):
        """Indicate whether the remote device is paired."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Paired')

    @property
    def connected(self):
        """Indicate whether the remote device is currently connected."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Connected')

    @property
    def trusted(self):
        """Indicate whether the remote device is seen as trusted."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE,
            'Trusted')

    @trusted.setter
    def trusted(self, new_state):
        """Indicate whether the remote device is seen as trusted."""
        self.remote_device_props.Set(
            constants.DEVICE_INTERFACE,
            'Trusted',
            new_state)

    @property
    def blocked(self):
        """Indicate whether the remote device is seen as blocked."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE,
            'Blocked')

    @blocked.setter
    def blocked(self, new_state):
        """Indicate whether the remote device is seen as blocked."""
        self.remote_device_props.Set(
            constants.DEVICE_INTERFACE,
            'Blocked',
            new_state)

    @property
    def alias(self):
        """remote device alias"""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Alias')

    @alias.setter
    def alias(self, new_alias):
        """remote device alias."""
        self.remote_device_props.Set(
            constants.DEVICE_INTERFACE,
            'Alias',
            new_alias)

    @property
    def _adapter(self):
        """The D-Bus object path of the adapter the device belongs to."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE, 'Adapter')

    @property
    def adapter(self):
        """The address of the adapter the device belongs to."""
        adapter_obj = dbus_tools.get_dbus_obj(self._adapter)
        adapter_props = dbus_tools.get_dbus_iface(dbus.PROPERTIES_IFACE,
                                                  adapter_obj)
        return dbus_tools.get(adapter_props,  constants.ADAPTER_INTERFACE,
                              'Address')

    @property
    def legacy_pairing(self):
        """Indicate the legacy pairing status.

        Set to true if the device only supports the pre-2.1 pairing mechanism.
        """
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE,
            'LegacyPairing')

    @legacy_pairing.setter
    def legacy_pairing(self, new_status):
        """Indicate the legacy pairing status.

        Set to true if the device only supports the pre-2.1 pairing mechanism.
        """
        self.remote_device_props.Set(
            constants.DEVICE_INTERFACE,
            'LegacyPairing',
            new_status)

    @property
    def modalias(self):
        """
        Remote Device ID information in modalias format.

        Used by the kernel and udev.
        """
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'Modalias', None)

    @property
    def RSSI(self):  # pylint: disable=invalid-name
        """
        Received Signal Strength Indicator of the remote device.

        (This is inquiry or advertising RSSI).
        """
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'RSSI', None)

    @property
    def tx_power(self):
        """Advertised transmitted power level (inquiry or advertising)."""
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'TxPower', None)

    @property
    def manufacturer_data(self):
        """
        Manufacturer specific advertisement data.

        Keys are 16 bits Manufacturer ID followed by its byte array value.
        """
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'ManufacturerData', None)

    @property
    def service_data(self):
        """
        Service advertisement data.

        Keys are the UUIDs in string format followed by its byte array value.
        """
        return dbus_tools.get(self.remote_device_props,
                              constants.DEVICE_INTERFACE,
                              'ServiceData', None)

    @property
    def services_resolved(self):
        """Indicate whether or not service discovery has been resolved."""
        return self.remote_device_props.Get(
            constants.DEVICE_INTERFACE,
            'ServicesResolved')

    @property
    def services_available(self):
        """Get a list of Service UUIDs available on this device"""
        return dbus_tools.get_services(self.remote_device_obj.object_path)

    def pair(self):
        """
        Pair the device
        """
        self.remote_device_methods.Pair()

    def cancel_pairing(self):
        """
        This method can be used to cancel a pairing
        operation initiated by the pair method
        """
        self.remote_device_methods.CancelPairing()

    def connect(self, profile=None, timeout=35):
        """
        Initiate a connection to the remote device.

        :param profile: (optional) profile to use for the connection.
        :param timeout: (optional) the number of seconds to spend trying
               to connect.
        """
        try:
            if profile is None:
                self.bus.call_blocking(constants.BLUEZ_SERVICE_NAME,
                                       self.remote_device_path,
                                       constants.DEVICE_INTERFACE,
                                       'Connect', '', [], timeout=timeout)
            else:
                # need a device to test next line with.  for now, timeout
                # ignored for profile connections.
                # self.bus.call_blocking(constants.BLUEZ_SERVICE_NAME,
                # self.remote_device_path, constants.DEVICE_INTERFACE,
                # 'ConnectProfile', 's', [profile], timeout=timeout)
                self.remote_device_methods.ConnectProfile(profile)
        except dbus.exceptions.DBusException as dbus_exception:
            dbus_error_type = 'org.freedesktop.DBus.Error.NoReply'
            if dbus_exception.get_dbus_name() == dbus_error_type:
                # move driver back from connecting state to disconnected state
                self.disconnect()
            raise dbus_exception

    def disconnect(self):
        """Disconnect from the remote device."""
        self.remote_device_methods.Disconnect()



"""
The level 10 file for creating beacons
This requires BlueZ to have the experimental flag set
"""
from bluezero import adapter
from bluezero import advertisement


class Beacon:
    """
    Create a non-connectable Bluetooth instance advertising information
    """
    def __init__(self, adapter_addr=None):
        """Default initialiser.

        Creates the BLE beacon object
        If an adapter object exists then give it as an optional argument
        If an adapter object is not given then the first adapter found is used

        :param adapter_addr: Optional Python adapter object.
        """
        self.dongle = None
        if adapter_addr is None:
            self.dongle = adapter.Adapter(adapter.list_adapters()[0])
        else:
            self.dongle = adapter.Adapter(adapter_addr)

        self.broadcaster = advertisement.Advertisement(1, 'broadcast')

    def add_service_data(self, service, data):
        """
        Add service and service data to be used in beacon message

        :param service: Valid service UUID
        :param data: Data to be sent (Limit of ??)
        """
        self.broadcaster.service_UUIDs = [service]
        self.broadcaster.service_data = {service: data}

    def add_manufacturer_data(self, manufacturer, data):
        """
        Add manufacturer information to be used in beacon message


        :param manufacturer: Use numbers from Bluetooth SIG
            https://www.bluetooth.com/specifications/assigned-numbers/16-bit-UUIDs-for-Members
        :param data: Data to be sent (Limit of 23bytes) e.g. b'\\xff' * 23
        """
        if isinstance(manufacturer, str):
            manufacturer = int(manufacturer, 16)
        self.broadcaster.manufacturer_data(manufacturer, data)

    def include_tx_power(self, show_power=None):
        """
        Use to include TX power in advertisement.
        This is different to the TX power in specific beacon format
        (e.g. Eddystone)

        :param show_power: boolean value
        :return:
        """
        if show_power is None:
            return self.broadcaster.include_tx_power
        else:
            self.broadcaster.include_tx_power = show_power

    def start_beacon(self):
        """
        Start beacon advertising
        """
        if not self.dongle.powered:
            self.dongle.powered = True
        ad_manager = advertisement.AdvertisingManager(self.dongle.address)
        ad_manager.register_advertisement(self.broadcaster, {})

        try:
            self.broadcaster.start()
        except KeyboardInterrupt:
            self.broadcaster.stop()
            ad_manager.unregister_advertisement(self.broadcaster)
        finally:
            pass



import dbus
import dbusmock
from pathlib import Path
import subprocess
from unittest import mock, skip
from tests.mock_eventloop import MockAsync, run_pending_events

from bluezero import adapter
from bluezero.adapter import AdapterError
from bluezero import observer


class TestBlueZ5(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')

    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_no_adapters(self):
        # Check for adapters.
        with self.assertRaises(AdapterError):
            out = adapter.list_adapters()

    def test_no_adapters_available(self):
        out = adapter.Adapter.available()
        with self.assertRaises(AdapterError):
            next(out)

    def test_one_adapter(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        dongle = adapter.Adapter()
        name = dongle.name
        self.assertEqual('My-Test-Device', name)

    def test_available_list(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        addresses = [dongle.address for dongle in adapter.Adapter.available()]
        self.assertEqual(['00:01:02:03:04:05'], addresses)

    def test_one_adapter_available(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        dongles = adapter.Adapter.available()
        for dongle in dongles:
            self.assertEqual('My-Test-Device', dongle.name)
            self.assertEqual('00:01:02:03:04:05', dongle.address)
            self.assertEqual('0x10c', hex(dongle.bt_class))
            self.assertEqual('My-Test-Device', dongle.alias)
            dongle.alias = 'New Alias'
            self.assertEqual('New Alias', dongle.alias)
            self.assertEqual(True, dongle.powered)

    def test_on_device_found(self):
        device_address = '11:01:02:03:04:05'

        class ForTest:
            found_address = None

            @classmethod
            def new_dev(cls, device):
                cls.found_address = device.address

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        dongle = adapter.Adapter()
        dongle.on_device_found = ForTest.new_dev
        self.dbusmock_bluez.AddDevice('hci0',
                                      device_address,
                                      'My-Peripheral-Device')
        run_pending_events()

        self.assertEqual(device_address, ForTest.found_address)

    def test_scanner_altbeacon(self):
        class ForTest:
            found_data = None

            @classmethod
            def new_dev(cls, device):
                cls.found_data = device.major

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            observer.Scanner.start_beacon_scan(on_altbeacon=ForTest.new_dev)
        self.assertEqual(24931, ForTest.found_data)

    @skip("Fails when run as part of suite but passses on own. Skipping for now")
    def test_scanner_iBeacon(self):
        class ForTest:
            found_data = None

            @classmethod
            def new_dev(cls, device):
                cls.found_data = device.major

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            observer.Scanner.start_beacon_scan(on_ibeacon=ForTest.new_dev)
        self.assertEqual(278, ForTest.found_data)

    def test_scanner_eddy_url(self):

        class ForTest:
            found_data = None

            @classmethod
            def new_dev(cls, device):
                cls.found_data = device.url

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            observer.Scanner.start_beacon_scan(
                on_eddystone_url=ForTest.new_dev)
        self.assertEqual('https://www.bluetooth.com',
                         ForTest.found_data)

    def test_scanner_eddy_url2(self):

        class ForTest:
            found_data = None

            @classmethod
            def new_dev(cls, device):
                cls.found_data = device.url

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            observer.scan_eddystone(
                on_data=ForTest.new_dev)
        self.assertEqual('https://www.bluetooth.com',
                         ForTest.found_data)

    def test_scanner_eddy_uid(self):

        class ForTest:
            found_data = None

            @classmethod
            def new_dev(cls, device):
                cls.found_data = device.namespace

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        with mock.patch('bluezero.async_tools.EventLoop.run', MockAsync.run):
            observer.Scanner.start_beacon_scan(
                on_eddystone_uid=ForTest.new_dev)
        self.assertEqual(297987634280,
                         ForTest.found_data)
